import json
import logging
from typing import Dict, Optional

from authlib.integrations.starlette_client import OAuth
from solara.server import settings
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    SimpleUser,
    UnauthenticatedUser,
)
from starlette.requests import HTTPConnection, Request
from starlette.responses import RedirectResponse

from .. import license

logger = logging.getLogger("solara.enterprise.auth.starlette")


oauth: Optional[OAuth] = None


def init():
    global oauth
    if settings.oauth.client_id:
        api_base_url = settings.oauth.api_base_url
        if not api_base_url.startswith("https://") and not api_base_url.startswith("http://"):
            api_base_url = f"https://{api_base_url}"
        oauth = OAuth()
        oauth.register(
            name="oauth1",
            client_id=settings.oauth.client_id,
            client_secret=settings.oauth.client_secret,
            api_base_url=api_base_url,
            server_metadata_url=f"{api_base_url}/.well-known/openid-configuration",
            client_kwargs={"scope": settings.oauth.scope},
        )


def check_oauth():
    assert oauth is not None
    assert oauth.oauth1 is not None
    if oauth.oauth1.client_id != settings.oauth.client_id:
        init()


init()


async def authorize(request: Request):
    check_oauth()
    assert oauth is not None
    assert oauth.oauth1 is not None

    org_url = request.session.pop("redirect_uri", settings.main.base_url + "/")

    token = await oauth.oauth1.authorize_access_token(request)
    # workaround: if token is set in the session in one piece, it is not saved, so we
    # split it up
    token.pop("id_token", None)
    user = token.pop("userinfo", None)
    request.session["token"] = json.dumps(token)
    request.session["user"] = json.dumps(user)

    return RedirectResponse(org_url)


async def logout(request: Request):
    redirect_uri = request.query_params.get("redirect_uri", "/")
    # ideally, we only remove these:
    request.session.pop("token", None)
    request.session.pop("user", None)
    request.session.pop("client_id", None)
    # but authlib sometimes leaves some stuff in the session on failed logins
    # so we clear it all
    request.session.clear()
    return RedirectResponse(redirect_uri)


async def login(request: Request, redirect_uri: Optional[str] = None):
    license.check("auth")
    check_oauth()
    assert oauth is not None
    assert oauth.oauth1 is not None
    if "redirect_uri" in request.query_params:
        # we arrived here via the auth.get_login_url() call, which means the
        # redirect_uri is in the query params
        request.session["redirect_uri"] = request.query_params["redirect_uri"]
    else:
        # otherwise we assume we got here via the solara.server.starlette method
        # where it detect we the OAuth.required=True setting, leading to a redirect
        request.session["redirect_uri"] = str(request.url.path)
    request.session["client_id"] = settings.oauth.client_id
    result = await oauth.oauth1.authorize_redirect(request, str(request.base_url) + "_solara/auth/authorize")
    return result


def get_user(request: HTTPConnection) -> Optional[Dict]:
    user = request.session.get("user")
    if user:
        user = json.loads(request.session["token"])
        user["userinfo"] = json.loads(request.session["user"])
        client_id = request.session.get("client_id")
        if client_id != settings.oauth.client_id:
            user = None
    return user


# only provides request.user.is_authenticated
class AuthBackend(AuthenticationBackend):
    async def authenticate(self, conn: HTTPConnection):
        user = get_user(conn)
        if user is None:
            return AuthCredentials(), UnauthenticatedUser()
        else:
            username = "noname"
            return AuthCredentials(["authenticated"]), SimpleUser(username)



import logging
import multiprocessing.pool
import threading
import time
import typing
from pathlib import Path
from typing import List, Optional

import solara
from rich import print as rprint
from solara.server import settings
from typing_extensions import TypedDict

from . import license

logger = logging.getLogger("solara.server.ssg")

if typing.TYPE_CHECKING:
    import playwright.sync_api
    import playwright.sync_api._context_manager


class Playwright(threading.local):
    browser: Optional["playwright.sync_api.Browser"] = None
    sync_playwright: Optional["playwright.sync_api.Playwright"] = None
    context_manager: Optional["playwright.sync_api._context_manager.PlaywrightContextManager"] = None
    page: Optional["playwright.sync_api.Page"] = None


pw = Playwright()
playwrights: List[Playwright] = []


class SSGData(TypedDict):
    title: str
    html: str
    styles: List[str]
    metas: List[str]


def _get_playwright():
    if hasattr(pw, "browser") and pw.browser is not None:
        return pw
    from playwright.sync_api import sync_playwright

    pw.context_manager = sync_playwright()
    pw.sync_playwright = pw.context_manager.start()

    pw.browser = pw.sync_playwright.chromium.launch(headless=not settings.ssg.headed)
    pw.page = pw.browser.new_page()
    return pw


def ssg_crawl(base_url: str):
    license.check("SSG")
    import solara.server.app
    import solara.server.kernel

    build_path = settings.ssg.build_path
    assert build_path is not None
    build_path.mkdir(exist_ok=True)

    app_script = solara.server.app.apps["__default__"]
    rprint(f"Building {app_script.name} at {build_path}")
    routes = app_script.routes

    # although in theory we should be able to run this with multiple threads
    # there are issues with uvloop:
    #  e.g.: "Racing with another loop to spawn a process."
    thread_pool = multiprocessing.pool.ThreadPool(1)

    results = []
    for route in routes:
        results.append(thread_pool.apply_async(ssg_crawl_route, [f"{base_url}/", route, build_path, thread_pool]))

    def wait(async_result):
        results = async_result.get()
        for result in results:
            wait(result)

    for result in results:
        wait(result)
    thread_pool.terminate()
    for pw in playwrights:
        assert pw.browser is not None
        assert pw.context_manager is not None
        pw.browser.close()
        pw.context_manager.stop()

    rprint("Done building SSG")


def ssg_crawl_route(base_url: str, route: solara.Route, build_path: Path, thread_pool: multiprocessing.pool.ThreadPool):
    # if route
    url = base_url + (route.path if route.path != "/" else "")
    if not route.children:
        rprint("Check SSG for URL", url)
        build_path.mkdir(exist_ok=True, parents=True)
        path = build_path / ("index.html" if route.path == "/" else route.path + ".html")
        stale = False
        pw = _get_playwright()
        page = pw.page
        if path.exists():
            if route.file is None:
                rprint(f"File corresponding to {url} is not found (route: {route})")
            else:
                assert route.file is not None
                stale = path.stat().st_mtime < route.file.stat().st_mtime
                if stale:
                    rprint(f"Path {path} is stale: mtime {path} is older than {route.file} mtime {route.file.stat().st_mtime}")
        if not path.exists() or stale:
            rprint(f"Will generate {path}")
            response = page.goto(url, wait_until="networkidle")
            if response.status != 200:
                raise Exception(f"Failed to load {url} with status {response.status}")
            # TODO: if we don't want to detached, we get stack trace showing errors in solara
            # make sure the html is loaded
            try:
                page.locator("#app").wait_for()
                # make sure vue took over
                page.locator("#pre-rendered-html-present").wait_for(state="detached")
                # and wait for the
                page.locator("text=Loading app").wait_for(state="detached")
                page.locator("#kernel-busy-indicator").wait_for(state="hidden")
                # page.wait_
                time.sleep(0.5)
                html = page.content()
            except Exception:
                logger.exception("Failure retrieving content for url: %s", url)
                raise
            path.write_text(html, encoding="utf-8")
            rprint(f"Wrote to {path}")
            page.goto("about:blank")
        else:
            rprint(f"Skipping existing render: {path}")
    results = []
    for child in route.children:
        result = thread_pool.apply_async(ssg_crawl_route, [url + "/", child, build_path / Path(route.path), thread_pool])
        results.append(result)
    return results


def ssg_data(path: str) -> Optional[SSGData]:
    license.check("SSG")
    html = ""
    # pre_rendered_css = ""
    styles = []
    title = "Solara ☀️"
    # still not sure why we sometimes end with a double slash
    if path.endswith("//"):
        path = path[:-2]
    if path.endswith("/"):
        path = path[:-1]
    if path.startswith("/"):
        # remove / so we don't get absolute paths on disk
        path = path[1:]
    # TODO: how do we know the app?
    build_path = settings.ssg.build_path
    if build_path and settings.ssg.enabled:
        html_path = build_path / path
        if (html_path / "index.html").exists():
            html_path = html_path / "index.html"
        else:
            html_path = html_path.with_suffix(".html")
        if html_path.exists() and html_path.is_file():
            logger.info("Using pre-rendered html at %r", html_path)

            from bs4 import BeautifulSoup, Tag

            soup = BeautifulSoup(html_path.read_text("utf8"), "html.parser")
            node = soup.find(id="app")
            # TODO: add classes...
            if node and isinstance(node, Tag):
                # only render children
                html = "".join(str(x) for x in node.contents)
            title_tag = soup.find("title")
            if title_tag:
                title = title_tag.text

            # include all meta tags
            rendered_metas = soup.find_all("meta")
            metas = []
            for meta in rendered_metas:
                # but only the ones added by solara
                if meta.attrs.get("data-solara-head-key"):
                    metas.append(str(meta))

            # include all styles
            rendered_styles = soup.find_all("style")
            for style in rendered_styles:
                style_html = str(style)
                # in case we want to skip the mathjax css
                # if "MJXZERO" in style_html:
                #     continue
                # pre_rendered_css += style_html
                styles.append(style_html)
                logger.debug("Include style (size is %r mb):\n\t%r", len(style_html) / 1024**2, style_html[:200])
            return SSGData(title=title, html=html, styles=styles, metas=metas)
        else:
            logger.error("Count not find html at %r", html_path)
    return None



import logging
import shutil
from typing import MutableMapping

import diskcache
import solara.settings
import solara.util

logger = logging.getLogger("solara-enterprise.cache.disk")


class Disk(MutableMapping):
    def __init__(
        self,
        clear=False,
        max_size=solara.settings.cache.disk_max_size,
        path=solara.settings.cache.path,
    ):
        # same as solara.cache.Memory
        eviction_policy = "least-recently-used"
        max_size = solara.util.parse_size(max_size)
        if clear:
            try:
                logger.debug("Clearing disk cache: %s", path)
                shutil.rmtree(path)
            except OSError:  # Windows wonkiness
                logger.exception(f"Error clearing disk cache: {path}")

        self.diskcache = diskcache.Cache(path, size_limit=max_size, eviction_policy=eviction_policy)

    def __getitem__(self, key):
        return self.diskcache[key]

    def __setitem__(self, key, value):
        self.diskcache[key] = value

    def __delitem__(self, key):
        del self.diskcache[key]

    def __iter__(self):
        return iter(self.diskcache)

    def __len__(self):
        return len(self.diskcache)
